# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114,C0116

from xsget import __version__


def test_version(script_runner):
    ret = script_runner("xstxt", "-V")
    assert f"xstxt {__version__}" in ret.stdout



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from xsget.book import Book
from xsget.chapter import Chapter


def test_book_dataclass_default_values():
    """Test Book dataclass with default values."""
    book = Book()
    assert book.title == ""
    assert book.authors == []
    assert book.chapters == []


def test_book_dataclass_custom_values():
    """Test Book dataclass with custom values."""
    chapter1 = Chapter(title="Chapter 1", content="Content of chapter 1")
    chapter2 = Chapter(title="Chapter 2", content="Content of chapter 2")
    book = Book(
        title="My Awesome Book",
        authors=["Author One", "Author Two"],
        chapters=[chapter1, chapter2],
    )
    assert book.title == "My Awesome Book"
    assert book.authors == ["Author One", "Author Two"]
    assert len(book.chapters) == 2
    assert book.chapters[0].title == "Chapter 1"
    assert book.chapters[1].title == "Chapter 2"



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Data model for an extracted Chapter."""

from dataclasses import dataclass, field


@dataclass(repr=False)
class Chapter:
    """A Chapter model class."""

    title: str = field(default="")
    content: str = field(default="")
    content_path: str = field(default="")
    filename: str = field(default="")

    def __repr__(self) -> str:
        return (
            f"Chapter(content='{self.content}', title='{self.title}', "
            f"filename='{self.filename}', content_path='{self.content_path}')"
        )

    def __str__(self) -> str:
        if self.title and self.content:
            return f"{self.title}\n\n{self.content}"
        if self.title:
            return self.title
        if self.content:
            return self.content
        return ""



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114,C0116

import argparse
from pathlib import Path

import aiohttp
import pytest
from aioresponses import aioresponses

from xsget.xsget import (
    extract_urls,
    fetch_url_by_aiohttp,
    http_headers,
    url_to_filename,
)

DEFAULT_URL = "http://localhost"


def test_url_to_filename():
    expected = [
        ("http://a.com", "index.html"),
        ("http://a.com/", "index.html"),
        ("http://a.com/123", "123.html"),
        ("http://a.com/123/456", "456.html"),
        ("http://a.com/123/456/789", "789.html"),
        ("http://a.com/123.html", "123.html"),
        ("http://a.com/123.html?abc=def", "123.html"),
    ]
    for url, filename in expected:
        assert url_to_filename(url) == filename

    expected = [
        ("http://a.com/123?id=aaa", "id", "aaa.html"),
        ("http://a.com/456.php?tid=abc", "tid", "abc.html"),
        ("http://a.com/789.php?test=xyz&id=aaa", "id", "aaa.html"),
    ]
    for url, url_param, filename in expected:
        assert url_to_filename(url, url_param) == filename


def test_extract_urls():
    html = """
        <html>
        <body>
        <div class="toc">
            <a href="http://a.com/123"/>a</a>
            <a href="http://a.com/123/789.html"/>b</a>
            <a href="//a.com/987"/>c</a>
            <a href="/123/456"/>d</a>
            <a href="/123/654.html"/>e</a>
        </div>
        </body>
        </html>
    """

    expected_urls = [
        "http://a.com/123",
        "http://a.com/123/789.html",
        "http://a.com/987",
        "http://a.com/123/456",
        "http://a.com/123/654.html",
    ]

    css_paths = [
        "html body div.toc a",
        "html body div a",
        "body div.toc a",
        "div.toc a",
        "div a",
        "a",
    ]
    for css_path in css_paths:
        config = argparse.Namespace(
            url="http://a.com/123", link_css_path=css_path
        )
        assert extract_urls(html, config) == expected_urls


def test_user_agent():
    user_agent = http_headers()["User-Agent"]
    assert "Mozilla/5.0" in user_agent


@pytest.mark.skip(reason="TODO")
async def test_fetch_url_by_aiohttp(tmpdir):
    session = aiohttp.ClientSession()
    with aioresponses() as mocked:
        config = argparse.Namespace(
            url_param_as_filename="",
            output_dir="output",
        )
        mocked.get(DEFAULT_URL, status=200, body="test")

        resp = await fetch_url_by_aiohttp(session, DEFAULT_URL, config)
        assert resp.status == 200
        mocked.assert_called_once_with(DEFAULT_URL)

        with open(
            Path(tmpdir, "output", "index.html"), encoding="utf8"
        ) as file:
            assert file.read() == "test"

        await session.close()



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Data model for an extracted Book."""

from dataclasses import dataclass, field

from xsget.chapter import Chapter


@dataclass
class Book:
    """Represents a book composed of multiple chapters.

    Attributes:
        title: The title of the book.
        authors: A list of authors for the book.
        chapters: A list of Chapter objects representing the book's content.
    """

    title: str = field(default="")
    authors: list[str] = field(default_factory=list)
    chapters: list[Chapter] = field(default_factory=list, repr=False)



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114,C0116

import argparse
import re
import shutil
from pathlib import Path

import pytest
from bs4 import BeautifulSoup

from xsget.chapter import Chapter
from xsget.xstxt import (
    extract_body,
    extract_chapter,
    extract_title,
    generate_book,
    generate_multiple_txt,
    generate_single_txt,
    get_html_files,
    search_and_replace,
    wrap,
)

# Taken from: https://zh.wikisource.org/wiki/詩經/關雎
CTEXT = """\
　　孔子論《詩》，以《關雎》為始。言太上者民之父母，后夫人之
行不侔乎天地，則無以奉神靈之统而理萬物之宜，故《詩》曰：「窈
窕淑女，君子好逑。」言能致其貞淑，不贰其操，情欲之感無介乎容
儀，晏私之意不形乎動静，夫然後可以配至尊而為宗廟主。此綱紀之
首、王教之端也。
"""


def test_get_html_files_sorted_in_natural_order(tmpdir):
    for name in ["200", "2", "100"]:
        tmpdir.join(f"{name}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 0, []) == [
        f"{path}/2.html",
        f"{path}/100.html",
        f"{path}/200.html",
    ]


def test_get_html_files_exclude_filter(tmpdir):
    for name in ["a", "2", "xyz"]:
        tmpdir.join(f"{name}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 0, [f"{path}/xyz.html"]) == [
        f"{path}/2.html",
        f"{path}/a.html",
    ]


def test_get_single_html_file(tmpdir):
    single = tmpdir.join("single.html")
    single.write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/single.html"], 0, []) == [
        f"{path}/single.html"
    ]


def test_get_html_files_by_limit(tmpdir):
    for i in range(0, 5):
        tmpdir.join(f"{i}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 3, []) == [
        f"{path}/0.html",
        f"{path}/1.html",
        f"{path}/2.html",
    ]


def test_get_no_html_files(caplog):
    path = "foobar/*.html"
    assert get_html_files([path], 0, []) == []
    assert f"No input files found in: {path}" in caplog.text


def test_get_html_files_with_non_existent_exclude(caplog):
    # Create some dummy files
    path = "temp_dir"
    Path(path).mkdir(exist_ok=True)
    Path(path, "file1.html").touch()
    Path(path, "file2.html").touch()

    # Call with a non-existent exclude pattern
    result = get_html_files(
        [f"{path}/*.html"], 0, [f"{path}/non_existent_*.html"]
    )
    assert len(result) == 2  # Both files should still be included
    assert (
        f"No exclude files found in: {path}/non_existent_*.html" in caplog.text
    )

    # Clean up
    shutil.rmtree(path)


async def test_extract_chapter():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">My Content</div>
        </body>
        </html>
    """
    config = argparse.Namespace(
        **{
            "title_css_path": "title",
            "body_css_path": "div#content",
            "html_replace": [],
            "debug": False,
        }
    )
    chapter = await extract_chapter(html, config)
    chapter.filename = "123.html"
    assert chapter.filename == "123.html"
    assert chapter.title == "My Title"
    assert chapter.content == "My Content"
    assert str(chapter) == "My Title\n\nMy Content"
    assert (
        repr(chapter)
        == "Chapter(content='My Content', title='My Title', filename='123.html', content_path='')"
    )


@pytest.mark.asyncio
async def test_extract_chapter_with_html_replace():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph1
            <br/><br />
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph2
            <br/>
            <br />
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph3
            </div>
        </body>
        </html>
    """
    config = argparse.Namespace(
        **{
            "title_css_path": "title",
            "body_css_path": "div#content",
            "html_replace": [
                ("<br/>", "11"),
                ("<br />", "22"),
                ("&nbsp;&nbsp;", "33"),
            ],
            "debug": False,
        }
    )
    chapter = await extract_chapter(html, config)

    match_regex = r"\n\s+".join(
        [
            "My Title\n\n",
            "3333Paragraph1",
            "1122",
            "3333Paragraph2",
            "11",
            "22",
            "3333Paragraph3",
        ]
    )
    assert re.match(match_regex, str(chapter))


async def test_extract_chapter_without_css_path():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">My Content</div>
        </body>
        </html>
    """
    config = argparse.Namespace(
        **{
            "title_css_path": None,
            "body_css_path": None,
            "html_replace": [],
            "debug": False,
        }
    )
    chapter = await extract_chapter(html, config)
    assert chapter.title == ""
    assert chapter.content == ""
    assert str(chapter) == ""
    assert (
        repr(chapter)
        == "Chapter(content='', title='', filename='', content_path='')"
    )


def test_generate_single_txt(tmpdir):
    chapters = [
        Chapter("MyTitle1", "MyContent1"),
        Chapter("MyTitle2", "MyContent2"),
    ]

    config = argparse.Namespace(
        **{
            "book_title": "Book Title",
            "book_author": "Book Author",
            "output": str(Path(tmpdir, "book.txt")),
            "txt_replace": (),
            "width": 60,
            "indent_chars": "",
            "paragraph_separator": "\n\n",
            "fullwidth": False,
            "language": "zh",
            "output_dir": "output",
        }
    )
    generate_single_txt(chapters, config)

    with open(config.output, encoding="utf8") as file:
        content = file.read()
        assert content == (
            "---\n"
            "书名：Book Title\n"
            "作者：Book Author\n"
            "---\n\n"
            "MyTitle1\n\nMyContent1\n\n"
            "MyTitle2\n\nMyContent2"
        )


def test_generate_single_txt_with_search_and_replace(tmpdir):
    book = [
        Chapter("MyTitle1", "MyContent1"),
        Chapter("MyTitle2", "MyContent2"),
    ]

    config = argparse.Namespace(
        **{
            "book_title": "Book Title",
            "book_author": "Book Author",
            "output": str(Path(tmpdir, "book.txt")),
            "txt_replace": [
                ("Title", "TITLE"),
                ("My", "YY"),
            ],
            "width": 60,
            "indent_chars": "",
            "paragraph_separator": "\n\n",
            "fullwidth": False,
            "language": "zh",
            "output_dir": "output",
        }
    )
    generate_single_txt(book, config)

    with open(config.output, encoding="utf8") as file:
        content = file.read()
        assert content == (
            "---\n"
            "书名：Book Title\n"
            "作者：Book Author\n"
            "---\n\n"
            "YYTITLE1\n\n"
            "YYContent1\n\n"
            "YYTITLE2\n\n"
            "YYContent2"
        )


def test_chapter_model_to_str():
    fixture_and_result = [
        (Chapter("MyTitle", "MyContent"), "MyTitle\n\nMyContent"),
        (Chapter("MyTitle", ""), "MyTitle"),
        (Chapter(title="MyTitle"), "MyTitle"),
        (Chapter("", "MyContent"), "MyContent"),
        (Chapter(content="MyContent"), "MyContent"),
        (Chapter(), ""),
    ]
    for fixture, result in fixture_and_result:
        assert str(fixture) == result


def test_search_and_replace_with_invalid_regex(caplog):
    search_and_replace("", [("[", "")])  # Invalid regex pattern
    assert "unterminated character set" in caplog.text


def test_search_and_replace_with_empty_regexs():
    content = "This is some content."
    result = search_and_replace(content, [])
    assert result == content


def test_wrap_by_width():
    config = argparse.Namespace(
        width=20,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert first_line == "\u3000\u3000孔子論《詩》，以"


def test_wrap_with_empty_content():
    config = argparse.Namespace(
        width=20,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
    )
    assert wrap("", config) == ""
    assert wrap("   ", config) == ""  # `wrap` strips whitespace for width > 0


def test_no_wrap():
    config = argparse.Namespace(
        width=0,
        indent_chars="",
        paragraph_separator="\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert (
        first_line
        == "\u3000\u3000孔子論《詩》，以《關雎》為始。言太上者民之父母，后夫人之"
    )


def test_wrap_with_indent_characters():
    config = argparse.Namespace(
        width=20,
        indent_chars="0000",
        paragraph_separator="\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert first_line == "0000孔子論《詩》"


async def test_extract_title_no_match():
    html = "<html><head><title>My Title</title></head><body></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    title = extract_title(soup, "h1")  # Non-existent CSS path
    assert title == ""


async def test_extract_title_empty_element():
    html = "<html><head><title></title></head><body></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    title = extract_title(
        soup, "title"
    )  # Existing CSS path, but empty element
    assert title == ""


async def test_extract_body_no_match():
    html = "<html><body><p>Content</p></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    body = extract_body(soup, "div#content")  # Non-existent CSS path
    assert body == ""


async def test_extract_body_empty_element():
    html = "<html><body><div id='content'></div></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    body = extract_body(
        soup, "div#content"
    )  # Existing CSS path, but empty element
    assert body == ""


def test_generate_multiple_txt(tmpdir):
    chapters = [
        Chapter("Title1", "Content1", filename="file1.html"),
        Chapter("Title2", "Content2", filename="file2.html"),
    ]
    output_dir = Path(tmpdir, "output_files")
    config = argparse.Namespace(
        output_dir=str(output_dir),
        txt_replace=[],
        fullwidth=False,
        width=0,
        indent_chars="",
        paragraph_separator="\n\n",
        language="en",
    )
    generate_multiple_txt(chapters, config)

    assert Path(output_dir, "file1.txt").exists()
    assert Path(output_dir, "file2.txt").exists()
    with open(Path(output_dir, "file1.txt"), encoding="utf8") as f:
        assert f.read() == "Title1\n\nContent1"
    with open(Path(output_dir, "file2.txt"), encoding="utf8") as f:
        assert f.read() == "Title2\n\nContent2"


@pytest.mark.asyncio
async def test_generate_book_purge(mocker, tmpdir):
    # Mock necessary functions to avoid actual file operations and async calls
    mocker.patch("xsget.xstxt.get_html_files", return_value=["dummy.html"])
    mocker.patch("xsget.xstxt.extract_chapter", return_value=Chapter("T", "C"))
    mocker.patch("xsget.xstxt.generate_single_txt")
    mocker.patch("xsget.xstxt.generate_multiple_txt")
    mocker.patch("shutil.rmtree")  # Mock rmtree

    # Mock aiofiles.open to prevent FileNotFoundError and support async with
    # Create a mock for the file handle that will be returned by aiofiles.open
    mock_file_handle = mocker.AsyncMock()
    mock_file_handle.read.return_value = b"<html><body>Content</body></html>"

    # Explicitly set __aenter__ and __aexit__ for the mock_file_handle
    # This is often not needed for AsyncMock, but can help debug stubborn issues.
    mock_file_handle.__aenter__ = mocker.AsyncMock(
        return_value=mock_file_handle
    )
    mock_file_handle.__aexit__ = mocker.AsyncMock(return_value=None)

    # Patch aiofiles.open to return this mock_file_handle directly.
    mocker.patch("aiofiles.open", return_value=mock_file_handle)

    output_dir = Path(tmpdir, "output")
    output_dir.mkdir()  # Create the directory to be purged

    config = argparse.Namespace(
        input=["*.html"],
        limit=0,
        exclude=[],
        debug=False,
        output_dir=str(output_dir),
        output="book.txt",
        purge=True,
        yes=False,  # Should still purge if purge is True and yes is False
        output_individual_file=False,
    )

    await generate_book(config)

    # Assert that rmtree was called with the correct path
    shutil.rmtree.assert_called_once_with(output_dir)



# Copyright (C) 2021,2022,2023,2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Console tools to download online novel and convert to text file."""

from datetime import datetime as dt
from importlib import metadata
from importlib.resources import read_text
from pathlib import Path
from typing import Any, Dict, cast
import argparse
import logging
import os
import sys

import tomlkit

__version__ = metadata.version("xsget")

_logger = logging.getLogger(__name__)


class ConfigFileCorruptedError(Exception):
    """Config file corrupted after reading."""


class ConfigFileExistsError(Exception):
    """Config file found when generating a new config file."""


def setup_logging(parsed_args: argparse.Namespace) -> None:
    """Set up logging based on the provided command-line arguments.

    Args:
        parsed_args (argparse.Namespace): Parsed command line arguments.
    """
    if parsed_args.quiet:
        logging.disable(logging.NOTSET)
        return

    level = logging.DEBUG if parsed_args.debug else logging.INFO
    format_string = (
        "[%(asctime)s] %(levelname)s: %(name)s: %(message)s"
        if parsed_args.debug
        else "%(message)s"
    )

    logging.basicConfig(
        level=level,
        format=format_string,
        stream=sys.stdout,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def load_or_create_config(
    parsed_args: argparse.Namespace, app: str
) -> Dict[str, Any]:
    """Load configuration from a file or create a new configuration file.

    Args:
        parsed_args: Parsed command line arguments.
        app: Application name to load config for.

    Returns:
        Dict containing merged configuration.

    Raises:
        ConfigFileCorruptedError: If config file is corrupted.
        ConfigFileExistsError: If generating config but file exists.
    """
    if parsed_args.config:
        return _load_config(parsed_args, app)

    if parsed_args.generate_config:
        return _create_config(parsed_args, app)

    return vars(parsed_args)


def _load_config(parsed_args: argparse.Namespace, app: str) -> Dict[str, Any]:
    """Load and validate configuration from a file.

    Args:
        parsed_args: Parsed command line arguments.
        app: Application name.

    Returns:
        Dict containing loaded configuration.

    Raises:
        ConfigFileCorruptedError: If config file is corrupted.
    """
    config_file = parsed_args.config

    with open(config_file, "r", encoding="utf8") as file:
        toml = tomlkit.load(file)

        if len(toml) == 0:
            raise ConfigFileCorruptedError(
                f"Corrupted config file: {config_file}"
            )

        toml_tpl = tomlkit.parse(read_text(__package__, f"{app}.toml"))
        if toml_tpl["config_version"] != toml.get("config_version"):
            _logger.info("Upgrade config file: %s", config_file)
            config = argparse.Namespace(**dict(toml))
            config.generate_config = config_file
            config.config_version = toml_tpl["config_version"]
            return _upgrade_config(config, app)

        _logger.info("Load from config file: %s", config_file)
        for key, value in toml.items():
            _logger.debug("config: %s, value: %s", repr(key), repr(value))

        return cast(Dict[str, Any], toml)


def _upgrade_config(config: argparse.Namespace, app: str) -> Dict[str, Any]:
    """Upgrade the configuration file to the latest version.

    Args:
        config: Parsed command line arguments.
        app: Application name.

    Returns:
        Dict containing the upgraded configuration.
    """
    toml_filename = Path(config.generate_config)

    ymd_hms = dt.now().strftime("%Y%m%d_%H%M%S")
    backup_filename = Path(
        toml_filename.resolve().parent.joinpath(
            toml_filename.stem + "_" + ymd_hms + "_backup.toml"
        )
    )
    os.rename(toml_filename, backup_filename)
    _logger.info("Backup config file: %s", backup_filename)

    return _create_config(config, app)


def _create_config(
    parsed_args: argparse.Namespace, app: str
) -> Dict[str, Any]:
    """Create a new configuration file with default values.

    Args:
        parsed_args: Parsed command line arguments.
        app: Application name.

    Returns:
        Dict containing the created configuration.

    Raises:
        ConfigFileExistsError: If a config file already exists.
    """
    config_filename = Path(parsed_args.generate_config)

    if config_filename.exists():
        raise ConfigFileExistsError(
            f"Existing config file found: {config_filename}"
        )

    with open(config_filename, "w", encoding="utf8") as file:
        config_dict = vars(parsed_args)
        _logger.debug(config_dict)

        toml = read_text(__package__, f"{app}.toml")
        doc = tomlkit.parse(toml)

        for key, value in config_dict.items():
            if key in doc:
                doc[key] = value

        file.write(tomlkit.dumps(doc))
        _logger.info("Create config file: %s", config_filename)

        return config_dict
